/*
 *  Render C declarations as English.
 *  Copyright © 2011 Nick Bowler
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>

#include "cdecl.h"
#include "typemap.h"

/* declare [ident] as [storage] [function specs] [other crap] [qualifiers] [type specs] */

static size_t
voutput(char *buf, size_t n, size_t off, const char *fmt, va_list ap)
{
	if (off >= n)
		return vsnprintf(NULL, 0, fmt, ap);
	return vsnprintf(buf+off, n-off, fmt, ap);
}

static size_t output(char *buf, size_t n, size_t off, const char *fmt, ...)
{
	va_list ap;
	size_t ret;

	va_start(ap, fmt);
	ret = voutput(buf, n, off, fmt, ap);
	va_end(ap);

	return ret;
}

static size_t advance_(char **buf, size_t *n, size_t amount)
{
	if (amount >= *n) {
		*n   = 0;
		*buf = 0;
	} else {
		*buf += amount;
		*n   -= amount;
	}

	return amount;
}

static size_t advance(char **buf, size_t *n, size_t amount)
{
	size_t ret, rc;

	if (!amount)
		return 0;

	ret = advance_(buf, n, amount);
	rc = snprintf(*buf, *n, " ");
	return ret + advance_(buf, n, rc);
}

static size_t
explain_qualifiers(char *buf, size_t n, struct cdecl_declspec *s)
{
	unsigned long qualmap = 0;
	size_t ret = 0, rc = 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		if (cdecl_spec_kind(c) != CDECL_SPEC_QUAL)
			continue;
		qualmap |= 1ul << (c->type & 0xff);
	}

	if (qualmap & (1ul << (CDECL_QUAL_RESTRICT & 0xff))) {
		ret += advance(&buf, &n, rc);
		rc = snprintf(buf, n, "restrict");
	}
	if (qualmap & (1ul << (CDECL_QUAL_VOLATILE & 0xff))) {
		ret += advance(&buf, &n, rc);
		rc = snprintf(buf, n, "volatile");
	}
	if (qualmap & (1ul << (CDECL_QUAL_CONST & 0xff))) {
		ret += advance(&buf, &n, rc);
		rc = snprintf(buf, n, "const");
	}

	return ret + rc;
}

/* Renders the type qualifiers and type specifiers in canonical form. */
static size_t
explain_post_specs(char *buf, size_t n, struct cdecl_declspec *s)
{
	const char *tag = NULL;
	unsigned long typemap;
	size_t ret = 0, rc;

	typemap = cdecl__build_typemap(s);
	if (typemap == -1)
		return 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		if (cdecl_spec_kind(c) != CDECL_SPEC_TYPE)
			continue;

		/* Valid C types have at most one identifier. */
		if (c->ident)
			tag = c->ident;
	}

	rc = explain_qualifiers(buf, n, s);
	ret += advance(&buf, &n, rc);

	rc = snprintf(buf, n, "%s", cdecl__explain_typemap(typemap));
	if (tag) {
		ret += advance(&buf, &n, rc);
		rc = snprintf(buf, n, "%s", tag);
	}

	return ret + rc;
}

static const char *explain_storage(unsigned spec)
{
	switch (spec) {
	case CDECL_STOR_TYPEDEF:
		return "typedef";
	case CDECL_STOR_EXTERN:
		return "extern";
	case CDECL_STOR_STATIC:
		return "static";
	case CDECL_STOR_AUTO:
		return "auto";
	case CDECL_STOR_REGISTER:
		return "register";
	default:
		assert(0);
	}
}

/* Renders the storage-class and function specifiers in canonical form. */
static size_t explain_pre_specs(char *buf, size_t n, struct cdecl_declspec *s)
{
	unsigned long funcmap = 0;
	const char *storage = NULL;
	size_t ret = 0;

	for (struct cdecl_declspec *c = s; c; c = c->next) {
		switch (cdecl_spec_kind(c)) {
		case CDECL_SPEC_FUNC:
			funcmap |= 1ul << (c->type & 0xff);
			break;
		case CDECL_SPEC_STOR:
			/* Valid C declarations have at most one
			 * storage-class specifier. */
			storage = explain_storage(c->type);
			break;
		}
	}

	if (storage)
		ret += output(buf, n, ret, "%s", storage);
	if (funcmap & (1ul << (CDECL_FUNC_INLINE & 0xff)))
		ret += output(buf, n, ret, "%.*s%s", !!ret, "", "inline");
	return ret;
}

/* Renders the name of the thing being declared. */
static size_t
explain_prologue(char *buf, size_t n, struct cdecl_declarator *d)
{
	while (d) {
		switch (d->type) {
		case CDECL_DECL_NULL:
			return snprintf(buf, n, "type");
		case CDECL_DECL_IDENT:
			return snprintf(buf, n, "declare %s as", d->u.ident);
		}

		d = d->child;
	}
}

static size_t
explain_pointer(char *buf, size_t n, struct cdecl_pointer *p)
{
	size_t ret = 0, rc;

	rc = explain_qualifiers(buf, n, p->qualifiers);
	ret += advance(&buf, &n, rc);

	return ret + snprintf(buf, n, "pointer to");
}

static size_t
explain_array(char *buf, size_t n, struct cdecl_array *a)
{
	size_t ret = 0, rc = 0;

	if (a->vla)
		rc = snprintf(buf, n, "variable-length array");
	else
		rc = snprintf(buf, n, "array");
	ret += advance(&buf, &n, rc);

	if (a->vla) {
		rc = snprintf(buf, n, "%s", a->vla);
		ret += advance(&buf, &n, rc);
	} else if (a->length) {
		rc = snprintf(buf, n, "%ju", a->length);
		ret += advance(&buf, &n, rc);
	}

	return ret + snprintf(buf, n, "of");
}

static size_t
explain_declarators(char *buf, size_t n, struct cdecl_declarator *d)
{
	size_t ret = 0, rc;

	if (d->type == CDECL_DECL_IDENT || d->type == CDECL_DECL_NULL)
		return 0;

	rc = explain_declarators(buf, n, d->child);
	ret += advance(&buf, &n, rc);

	switch (d->type) {
	case CDECL_DECL_POINTER:
		return ret + explain_pointer(buf, n, &d->u.pointer);
	case CDECL_DECL_ARRAY:
		return ret + explain_array(buf, n, &d->u.array);
	default:
		assert(0);
	}
}

size_t cdecl_explain(char *buf, size_t n, struct cdecl *decl)
{
	size_t ret = 0, rc;

	rc = explain_prologue(buf, n, decl->declarators);
	ret += advance(&buf, &n, rc);

	rc = explain_pre_specs(buf, n, decl->specifiers);
	ret += advance(&buf, &n, rc);

	rc = explain_declarators(buf, n, decl->declarators);
	ret += advance(&buf, &n, rc);

	return ret + explain_post_specs(buf, n, decl->specifiers);
}
